module.exports = {
    "env": {
        "browser": true,
        "jquery": true,
        "node": true,
        "es2021": true
    },
    "extends": [
        "eslint:recommended"
    ],
    "parserOptions": {
        "sourceType": "module"
    },
    "rules": {
    }
};
